/*
 * spi.c
 *
 *  Created on: 21 Aug 2011
 *      Author: Mike
 */

#include "spi.h"

void spiDoIdle(void)
{
	// service stuff we have to do to keep the world spinning
	//=======================================================
	boardService();
	usbComService();
}

uint8 spiDeselect(uint8 usartID)
{
	// deselect the nominated SPI device by setting SELECT line HIGH
	// returns PORT value
	// returns 0 on error
	//==============================================================

	uint8 alternate;
	uint8 port;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiDeselect USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	if (usartID == 0)
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 1;

		// de-select the device
		//=====================
		if (alternate == 0)
		{
			// port 0 bit 4
			//=============
			port = P0;
			// debug("spiDeselect USART0 alternate 1 P0: %x\n", port);

			P0 |= 0x10;
			port = P0;
			// debug("spiDeselect USART0 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 2
			//=============
			port = P1;
			// debug("spiDeselect USART0 alternate 2 P1: %x\n", port);

			P1 |= 0x04;
			port = P1;
			// debug("spiDeselect USART0 alternate 2 P1: %x\n", port);
		}
	}
	else
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 2;

		// de-select the device
		//=====================
		if (alternate == 0)
		{
			// port 0 bit 2
			//=============
			port = P0;
			// debug("spiDeselect USART1 alternate 1 P0: %x\n", port);

			P0 |= 0x04;
			port = P0;
			// debug("spiDeselect USART1 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 4
			//=============
			port = P1;
			// debug("spiDeselect USART1 alternate 2 P1: %x\n", port);

			P1 |= 0x10;
			port = P1;
			// debug("spiDeselect USART1 alternate 2 P1: %x\n", port);
		}
	}
	return port;
}

uint8 spiInit(uint8 usartID, uint8 alternate, uint8 polarity, uint8 phase, uint8 bitOrder)
{
	// initialise SPI interface
	// return 1 for success
	// return 0 on error
	//=========================

	// USART_ID is 0 or 1
	// alternate is 1 or 2 for the two choices of alternative location
	// polarity is 0 for Negative, 1 for Positive
	// phase is 0 for Negative 1 for Positive
	// bitOrder is 0 for MSB first, 1 for LSB first
	// defines for easier reading:
	//
	//	spiUsart0 0
	//	spiUsart1 1
	//
	//	spiAlternate1 1
	//	spiAlternate2 2
	//
	//	spiPhaseN 0
	//	spiPhaseP 1
	//
	//	spiPolarityN 0
	//	spiPolarityP 1
	//
	//	spiBitOrderM 0
	//	spiBitOrderL 1
	//================================================================

	// range check usartID
	//====================
	if (usartID > spiUsart1)
	{
		debug("spiInit USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// range check alternate
	//=====================
	if (alternate < spiAlternate1 || alternate > spiAlternate2)
	{
		debug("alternate %d out of range[1-2]\n", alternate);
		return 0;
	}

	// range check polarity
	//=====================
	if (polarity > spiPolarityP)
	{
		debug("polarity %d out of range[0-1]\n", polarity);
		return 0;
	}

	// range check phase
	//==================
	if (phase > spiPhaseP)
	{
		debug("phase %d out of range[0-1]\n", phase);
		return 0;
	}

	// range check bitOrder
	//=====================
	if (bitOrder > spiBitOrderL)
	{
		debug("bitOrder %d out of range[0-1]\n", bitOrder);
		return 0;
	}

	if (usartID == spiUsart0)
	{
		// process USART0
		//===============
		debug("USART 0\n");

		// start by clearing U0CSR
		//========================
		U0CSR = 0;

		// set SPI mode
		//=============
		U0CSR &= ~0x80;	// U0CSR bit 7=0 is SPI mode
		debug("spiInit SPI mode U0CSR.MODE: %2x\n", U0CSR & 0x80);

		// set Master Mode
		//================
		U0CSR &= ~0x20;	// U0CSR bit 5=0 is Master mode
		debug("spiInit Master mode U0CSR: %2x\n", U0CSR & 0x20);

		if (alternate == spiAlternate1)
		{
			// set USART0 alternate 1 pins
			//============================
			PERCFG &= ~0x01;	// USART0 alternate 1
			P0SEL |= 0x2C;  	// P0_5, P0_3, and P0_2 are peripherals
			P0SEL &= ~0x10; 	// P0_4 is GPIO (SSN)
			P0DIR |= 0x10;  	// SSN is set as output

			debug("spiInit alternate 1 PERCFG.U0CFG: %2x\n", PERCFG & 0x01);
			debug("spiInit alternate 1 P0SEL: %2x\n", P0SEL);
			debug("spiInit alternate 1 P0DIR: %2x\n", P0DIR);
		}
		else
		{
			// set USART0 alternate 2 pins
			//============================
			PERCFG |= 0x01;		// USART0 alternate 2
			P1SEL |= 0x38;  	// P1_3, P1_4, and P1_5 are peripherals
			P1SEL &= ~0x04; 	// P1_2 is GPIO (SSN)
			P1DIR |= 0x04;  	// SSN is set as output

			debug("spiInit alternate 2 PERCFG.u0CFG: %2x\n", PERCFG & 0x01);
			debug("spiInit alternate 2 P1SEL: %2x\n", P1SEL);
			debug("spiInit alternate 2 P1DIR: %2x\n", P1DIR);
		}

		// Set baud rate to max (system clock frequency / 8)
		//==================================================
		spiSetSpeed(usartID, 0);
		debug("spiInit U0BAUD_M: %2x\n", U0BAUD);

		// SPI polarity
		//=============
		if (polarity == spiPolarityP)
			U0GCR |= 0x80;
		else
			U0GCR &= ~0x80;

		// SPI phase
		//==========
		if (phase == spiPhaseP)
			U0GCR |= 0x40;
		else
			U0GCR &= ~0x40;

		// SPI bit order
		//==============
		if (bitOrder == spiBitOrderM)
			U0GCR |= 0x20;
		else
			U0GCR &= ~0x20;

		debug("spiInit U0GCR.CPOLR:  %2x\n", U0GCR & 0x80);
		debug("spiInit U0GCR.CPHA:   %2x\n", U0GCR & 0x40);
		debug("spiInit U0GCR.ORDER:  %2x\n", U0GCR & 0x20);
		debug("spiInit U0GCR.BAUD_E: %2x\n", U0GCR & 0x1F);

		// priority USART 0
		//=================
		P2DIR &= ~0xC0;
		debug("spiInit P2DIR.PRIP0: %2x\n", P2DIR & 0xC0);
	}
	else
	{
		// process USART1
		//===============

		debug("USART1\n");

		// start by clearing U1CSR
		//========================
		U1CSR = 0;

		// set SPI mode
		//=============
		U1CSR &= ~0x80;	// U1CSR bit 7=0 is SPI mode

		debug("spiInit SPI mode U1CSR: %2x\n", U1CSR & 0x80);

		// set Master Mode
		//================
		U1CSR &= ~0x20;	// U1CSR bit 5=0 is Master mode

		debug("spiInit Master mode U1CSR: %2x\n", U1CSR & 0x20);

		if (alternate == spiAlternate1)
		{
			// set USART1 alternate 1 pins
			//============================
			PERCFG &= ~0x02;	// USART1 alternate 1
			P0SEL |= 0x38;  	// P0_5, P0_4, and P0_3 are peripherals
			P0SEL &= ~0x04; 	// P0_2 is GPIO (SSN)
			P0DIR |= 0x04;  	// SSN is set as output

			debug("spiInit alternate 1 PERCFG: %2x\n", PERCFG & 0x02);
			debug("spiInit alternate 1 P0SEL: %2x\n", P0SEL);
			debug("spiInit alternate 1 P0DIR: %2x\n", P0DIR);
		}
		else
		{
			// set USART1 alternate 2 pins
			//============================
			PERCFG |= 0x02;		// USART1 alternate 2
			P1SEL |= 0xE0;  	// P1_7, P1_6, and P1_5 are peripherals
			P1SEL &= ~0x10; 	// P1_4 is GPIO (SSN)
			P1DIR |= 0x10;  	// SSN is set as output

			debug("spiInit alternate 2 PERCFG: %2x\n", PERCFG & 0x02);
			debug("spiInit alternate 2 P1SEL: %2x\n", P1SEL);
			debug("spiInit alternate 2 P1DIR: %2x\n", P1DIR);
		}

		// Set baud rate to max (system clock frequency / 8)
		//==================================================
		spiSetSpeed(usartID, 0);
		debug("spiInit U1BAUD_M: %2x\n", U1BAUD);

		// SPI polarity
		//=============
		if (polarity == spiPolarityP)
			U1GCR |= 0x80;
		else
			U1GCR &= ~0x80;

		// SPI phase
		//==========
		if (phase == spiPhaseP)
			U1GCR |= 0x40;
		else
			U1GCR &= ~0x40;

		// SPI bit order
		//==============
		if (bitOrder == spiBitOrderM)
			U1GCR |= 0x20;
		else
			U1GCR &= ~0x20;

		debug("spiInit U1GCR.CPOLR:  %2x\n", U1GCR & 0x80);
		debug("spiInit U1GCR.CPHA:   %2x\n", U1GCR & 0x40);
		debug("spiInit U1GCR.ORDER:  %2x\n", U1GCR & 0x20);
		debug("spiInit U1GCR.BAUD_E: %2x\n", U1GCR & 0x1F);

		//1st priority: USART 1
		//=====================
		P2DIR &= ~0xC0;
		P2DIR |= 0x40;
		debug("spiInit P2DIR.PRIP0: %2x\n", P2DIR & 0xC0);
	}

	return 1;
}

uint8 spiReadByte(uint8 usartID)
{
	// read one byte of data from SPI
	// return byte read
	// return 0 on error
	//===============================

	uint8 buffer[8];
	uint16 size;

	// it's like spiReadData, but only one byte into private buffer
	//=============================================================

	// send command over SPI interface
	// it's like send data, but without the buffer
	// so we fake it!
	//================================

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiReadByte USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// read the data into our private buffer
	// then return it
	//======================================
	buffer[0] = 0xA5;
	size = spiReadData(usartID, buffer, 1);
	if (size == 1)
		return buffer[0];
	else
		return 0;
}

uint16 spiReadData(uint8 usartID, uint8* buffer, uint16 size)
{
	// read data from SPI
	// return amount of data read
	// return 0 on error
	//===========================

	uint16 i;
	uint32 startTime;
	uint32 timeSpent;
	uint32 idleCount;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiReadData USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// select the device
	//==================
	// spiSelect(usartID);

	// note current (ms) time
	//=======================
	startTime = getMs();
	idleCount = 0;

	// read data from SPI into buffer
	//===============================
	if (usartID == spiUsart0)
	{
		for (i = 0; i < size; i++)
		{
			// write empty data to the device
			//===============================
			U0DBUF = 0xFF;

			// wait for data receive flag
			//===========================
			while ((U0CSR | 0x04) == 0)
			{
				idleCount++;
				spiDoIdle();
				timeSpent = getMs() - startTime;
				if (timeSpent > TIMEOUT)
				{
					// took too long to respond
					//=========================
					debug("spiReadData timeout\n");
					debug("spiReadData idleCount %ld\n", idleCount);
					return 0;
				}
			}

			// read data into the buffer
			//==========================
			buffer[i] = U0DBUF;
		}
	}
	else
	{
		for (i = 0; i < size; i++)
		{
			// write empty data to the device
			//===============================
			U1DBUF = 0xFF;
			idleCount = 0;

			// wait for data receive flag
			//===========================
			while ((U1CSR | 0x04) == 0)
			{
				idleCount++;
				spiDoIdle();
				timeSpent = getMs() - startTime;
				if (timeSpent > TIMEOUT)
				{
					// took too long to respond
					//=========================
					debug("spiReadData timeout\n");
					debug("spiReadData idleCount %ld\n", idleCount);
					return 0;
				}
			}

			// read data into the buffer
			//==========================
			buffer[i] = U1DBUF;
		}
	}

	// deselect the device
	//====================
	// spiDeselect(usartID);

	// debug("spiReadData idleCount %ld\n", idleCount);
	return size;
}

uint8 spiSelect(uint8 usartID)
{
	// select the nominated SPI device by setting SELECT line HIGH
	// return PORT contents if ok
	// return 0 on error
	//============================================================

	uint8 alternate;
	uint8 port;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiSelect USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	if (usartID == 0)
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 1;

		// select the device
		//==================
		if (alternate == 0)
		{
			// port 0 bit 4
			//=============
			port = P0;
			// debug("spiSelect USART0 alternate 1 P0: %x\n", port);

			P0 &= ~0x10;
			port = P0;
			// debug("spiSelect USART0 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 2
			//=============
			port = P1;
			// debug("spiSelect USART0 alternate 2 P1: %x\n", port);

			P1 &= ~0x04;
			port = P1;
			// debug("spiSelect USART0 alternate 2 P1: %x\n", port);
		}
	}
	else
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 2;

		// select the device
		//==================
		if (alternate == 0)
		{
			// port 0 bit 2
			//=============
			port = P0;
			// debug("spiSelect USART1 alternate 1 P0: %x\n", port);

			P0 &= ~0x04;
			port = P0;
			// debug("spiSelect USART1 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 4
			//=============
			port = P1;
			// debug("spiSelect USART1 alternate 2 P1: %x\n", port);

			P1 &= ~0x10;
			port = P1;
			// debug("spiSelect USART1 alternate 2 P1: %x\n", port);
		}
	}
	return port;
}

uint16 spiSetSpeed(uint8 usartID, uint16 speed)
{
	// set SPI speed
	// return speed set
	// return 0 on error
	//==================

	uint8 E;
	uint8 M;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiSetSpeed USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	switch (speed)
	{
		case 0:
			// set maximum speed
			//==================
			M = 0;
			E = 17;
			break;
		case 200:
			// 200 Khz
			// actual is 199.95 KHz
			// error of -.025%
			// based on 24 MHz clock
			//======================
			M = 17;
			E = 13;
			break;
		case 400:
			// 400 Khz
			// actual is 399.90 KHz
			// error of -.025%
			// based on 24 MHz clock
			//======================
			M = 17;
			E = 14;
			break;
		default:
			// invalid speed
			//==============
			debug("spiSetSpeed speed %d not supported\n", speed);
			return 0;
			break;
	}

	if (usartID == 0)
	{
		// set USART0 speed
		//=================
		U0BAUD = M;
		U0GCR = E;
	}
	else
	{
		// set USART1 speed
		//=================
		U1BAUD = M;
		U1GCR = E;
	}
	return speed;
}

void spiShowBuffer( uint8 *buffer, uint16 size)
{
	uint16 i, r, c;

	// print the buffer block
	//=======================
	r = 0;
	c = 0;
	do
	{
		if (size-r*16 < 16)
			i = size-r*16;
		else
			i = 16;

		debug("%04X: ", r*16);
		for (c=0; c<i; c++)
			debug("%02X.", buffer[r*16+c]);
		debug("\n");
		r++;
	} while (r*16 < size);
}

uint16 spiWriteData(uint8 usartID, uint8* buffer, uint16 size)
{
	// write buffer full of data to SPI device
	// return amount of data written
	// return 0 on error
	//========================================
	uint16 i;
	uint32 startTime;
	uint32 timeSpent;
	uint32 tries;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiWriteData USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// dump buffer contents
	//=====================
	debug("spiWriteData USART%d %d bytes\n", usartID, size);
	spiShowBuffer(buffer, size);

	// select the device
	//==================
	// spiSelect(usartID);

	// note current (ms) time
	//=======================
	startTime = getMs();
	tries = 0;

	// write data from buffer to SPI
	//==============================
	if (usartID == spiUsart0)
	{
		// loop writing data
		//==================
		for (i = 0; i <= size; i++)
		{
			// clear data transmitted flag
			//============================
			U0CSR &= ~0x02;

			// send a byte
			//============
			U0DBUF = buffer[i];

			// wait for data transmitted
			//==========================
			while ((U0CSR | 0x02) == 0)
			{
				tries++;
				spiDoIdle();
				{
					timeSpent = getMs() - startTime;
					if (timeSpent > TIMEOUT)
					{
						// took too long to respond
						//=========================
						debug("spiWriteData timeout\n");
						return 0;
					}
				}
			}
		}
		// need to send a final 0xFF to let card complete
		//===============================================
		U0DBUF = 0xFF;
	}
	else
	{
		// loop writing data
		//==================
		for (i = 0; i <= size; i++)
		{
			// clear data transmitted flag
			//============================
			U1CSR &= ~0x02;

			// send a byte
			//============
			U1DBUF = buffer[i];

			// wait for data transmitted
			//==========================
			while ((U1CSR | 0x02) == 0)
			{
				tries++;
				spiDoIdle();
				timeSpent = getMs() - startTime;
				if (timeSpent > TIMEOUT)
				{
					// took too long to respond
					//=========================
					debug("spiWriteData timeout\n");
					return 0;
				}
			}
		}
		// need to send a final 0xFF to let card complete
		//===============================================
		U1DBUF = 0xFF;
	}

	// debug("spWriteData tries: %ld\n", tries);

	// de-select the device
	//=====================
	// spiDeselect(usartID);

	return size;
}
